
from generallibrary import Timer, EmptyContext


class _Lock:
    """ A one-time-use lock used by Path.lock.
        Creates a lock for folder or path with these steps:
            Wait until unlocked.
            Create lock.
            Make sure only locked by self.
        A lock is inactive if it can be removed, as there's no Lock holding it's file stream.
        Todo: Lock the optional extra paths.
        """
    def __init__(self, abs_path, *other_paths):
        self.path = abs_path
        self.lock_file_stream = None

    def __enter__(self):
        self._attempt_lock_creation()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._close_and_remove_lock()

    def _delete_if_dead(self, lock_path):
        try:
            seconds_since_creation = lock_path.seconds_since_modified()
        except (AttributeError, FileNotFoundError):
            pass
        else:
            if seconds_since_creation > lock_path.dead_lock_seconds:
                lock_path.delete(error=False)

    def _attempt_lock_creation(self):
        timer = Timer()
        while timer.seconds() < self.path.timeout_seconds:
            affecting_locks = list(self._affecting_locks())
            if self._is_locked_externally(affecting_locks=affecting_locks):
                for locked_path in affecting_locks:
                    self._delete_if_dead(locked_path.get_lock_path())
            else:
                if not self._open_and_create_lock():
                    continue  # If lock failed creating

                affecting_locks = list(self._affecting_locks())
                if self.path in affecting_locks:
                    if self._is_locked_externally(affecting_locks=affecting_locks):
                        self._close_and_remove_lock()  # Remove and try again to respect other locks
                    else:
                        return  # Success
                else:
                    raise FileNotFoundError(f"Lock '{self.path}' failed to create.")
        raise TimeoutError(f"Couldn't lock '{self.path}' in time.")

    def _open_and_create_lock(self):
        if self.lock_file_stream is not None:
            raise AttributeError(f"A file stream is already opened for '{self.path}'.")

        try:
            self.lock_file_stream = open(self.path.get_lock_path(), "x")
        except FileExistsError:
            return False

        self.lock_file_stream.write("foo")

        self.path.locked_paths.append(self.path)
        return True

    def _close_and_remove_lock(self):
        if self.lock_file_stream is None:
            raise AttributeError(f"A file stream is not opened for '{self.path}'.")

        self.path.locked_paths.remove(self.path)

        self.lock_file_stream.close()
        self.lock_file_stream = None

        self.path.get_lock_path().delete(error=False)

    def _affecting_locks(self):
        """ Returns absolute paths in list pointing to path it's locking.
            Use `get_alternative_path()` or 'get_lock_path()' if actual lock path is wanted.

            :rtype: tuple[generalfile.Path] """
        # print(len(self.path.get_lock_dir().get_children()))
        for alternative_path in self.path.get_lock_dir().get_children(gen=True):
            path = alternative_path.from_alternative()
            if self.path.startswith(path) or path.startswith(self.path):
                yield path

    def _is_locked_externally(self, affecting_locks=None):
        """ See if this lock's path is locked by an external process. """
        if affecting_locks is None:
            affecting_locks = self._affecting_locks()

        for path in affecting_locks:
            if path not in self.path.locked_paths:
                return True
        return False


class Path_ContextManager:
    """ Context manager methods for Path. """
    locked_paths = []

    @staticmethod
    def _create_context_manager(path, *other_paths):
        """ :param generalfile.Path path: """
        path_is_lock = path.startswith(path.get_lock_dir())
        path_already_locked_by_this_process = path in path.locked_paths
        if path_is_lock or path_already_locked_by_this_process:
            return EmptyContext()  # Create fake lock
        else:
            return _Lock(path, *other_paths)  # Create real lock

    def lock(self, *other_paths):
        """ Create a lock for this path unless path is inside `lock dir`.
            Optionally supply additional paths to create locks for them too.
            Locks are stored in a list Path's cls, so only locks outside of this scope are respected.
            That allows us to automatically create fake recursive locks, where all obey the first outer lock.

            :param generalfile.Path self: """
        return self._create_context_manager(self.absolute(), *other_paths)

    def as_working_dir(self):
        """ Temporarily set working dir.

            :param generalfile.Path self: """
        return TempWorkingDir(self)


class TempWorkingDir:
    """ Temporarily set working dir. """
    def __init__(self, path):
        self.path = path
        self.original_working_dir = path.get_working_dir()

    def __enter__(self):
        self.path.set_working_dir()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.original_working_dir.set_working_dir()































import pathlib
import os


from generallibrary import VerInfo, TreeDiagram, Recycle, classproperty, deco_cache

from generalfile.errors import InvalidCharacterError
from generalfile.path_lock import Path_ContextManager
from generalfile.path_operations import Path_Operations
from generalfile.path_strings import Path_Strings
from generalfile.optional_dependencies.path_spreadsheet import Path_Spreadsheet
from generalfile.optional_dependencies.path_text import Path_Text
from generalfile.optional_dependencies.path_cfg import Path_Cfg
from generalfile.optional_dependencies.path_pickle import Path_Pickle


class Path(TreeDiagram, Recycle, Path_ContextManager, Path_Operations, Path_Strings, Path_Spreadsheet, Path_Text, Path_Cfg, Path_Pickle):
    """ Immutable cross-platform Path.
        Built on pathlib and TreeDiagram.
        Implements rules to ensure cross-platform compatability.
        Adds useful methods.
        Todo: Binary extension. """
    verInfo = VerInfo()
    _path_delimiter = verInfo.pathDelimiter
    Path = ...

    _recycle_keys = {"path": lambda path: Path.scrub("" if path is None else str(path))}
    _alternative_chars = {_path_delimiter: "&#47;", ":": "&#58", ".": "&#46;"}

    def __init__(self, path=None):  # Don't have parent here because of Recycle
        self.path = self.scrub(str_path="" if path is None else str(path))

        self._path = pathlib.Path(self.path)
        self._latest_listdir = set()

    copy_node = NotImplemented  # Maybe something like this to disable certain methods

    @classproperty
    def path_delimiter(cls):
        return cls._path_delimiter

    def spawn_parents(self):
        if not self.get_parent(spawn=False) and self.path and not self.is_root():
            try:
                index = self.path.rindex(self.path_delimiter) + 1
            except ValueError:
                index = 0
            self.set_parent(Path(path=self.path[:index]))

    def spawn_children(self):
        if self.is_folder():
            old_children = {path.name() for path in self.get_children(spawn=False)}

            try:
                new_children = set(os.listdir(self.path if self.path else "."))
            except PermissionError:
                new_children = set()

            for name in old_children.symmetric_difference(new_children):
                path = Path(path=self / name)
                path.set_parent(self if name in new_children else None)

    def __str__(self):
        return getattr(self, "path", "<Path not loaded yet>")
        # return self.path

    def __repr__(self):
        return self.name()

    def __fspath__(self):
        return self.path

    def __format__(self, format_spec):
        return self.path.__format__(format_spec)

    def __truediv__(self, other):
        """ :rtype: generalfile.Path """
        # print("here", self._recycle_instances)
        return self.Path(self._path / str(other))

    def __eq__(self, other):
        if isinstance(other, Path):
            other = other.path
        else:
            other = self._scrub("" if other is None else str(other))
        return self.path == other

    def __hash__(self):
        return hash(self.path)

    def __contains__(self, item):
        return self.path.__contains__(item)

    @classmethod
    def _scrub(cls, str_path):
        str_path = cls._replace_delimiters(str_path=str_path)
        str_path = cls._invalid_characters(str_path=str_path)
        str_path = cls._trim(str_path=str_path)
        str_path = cls._delimiter_suffix_if_root(str_path=str_path)
        return str_path

    @classmethod
    @deco_cache()
    def scrub(cls, str_path):
        return cls._scrub(str_path=str_path)

    @classmethod
    @deco_cache()
    def _replace_delimiters(cls, str_path):
        str_path = str_path.replace("/", cls.path_delimiter)
        str_path = str_path.replace("\\", cls.path_delimiter)
        return str_path

    @classmethod
    @deco_cache()
    def _invalid_characters(cls, str_path):
        # Simple invalid characters testing from Windows
        for character in '<>"|?*':
            if character in str_path:
                raise InvalidCharacterError(f"Invalid character '{character}' in '{str_path}'")

        if ":" in str_path:
            if not cls.verInfo.pathRootHasColon:
                raise InvalidCharacterError(f"Path has a colon but '{cls.verInfo.os}' doesn't use colon for path root: '{str_path}'")
            if str_path[1] != ":":
                raise InvalidCharacterError(f"Path has a colon but there's no colon at index 1: '{str_path}'")
            if len(str_path) >= 3 and str_path[2] != cls.path_delimiter:
                raise InvalidCharacterError(f"Path has a colon but index 2 is not a delimiter: '{str_path}'")
            if ":" in str_path[2:]:
                raise InvalidCharacterError(f"Path has a colon that's not at index 1: '{str_path}'")

        if str_path.endswith("."):
            raise InvalidCharacterError(f"Path cannot end with a dot ('.').")
        return str_path

    @classmethod
    @deco_cache()
    def _trim(cls, str_path):
        if not cls.verInfo.pathRootIsDelimiter and str_path.startswith(cls.path_delimiter):
            str_path = str_path[1:]
        if str_path.endswith(cls.path_delimiter) and len(str_path) > 1:
            str_path = str_path[0:-1]
        return str_path

    @classmethod
    @deco_cache()
    def _delimiter_suffix_if_root(cls, str_path):
        if len(str_path) == 2 and str_path[1] == ":":
            return f"{str_path}{cls.path_delimiter}"
        return str_path

setattr(Path, "Path", Path)


















































from generallibrary import deco_cache, initBases
from generalfile.optional_dependencies._extension import _Extension

import re


class Path_Text:
    """ Text methods for Path. """
    @property
    @deco_cache()
    def text(self):
        """ Easily modify file by treating it as a string. """
        return _Text(self)


@initBases
class _Text(_Extension):
    def write(self, text=None, overwrite=False):
        """ Write to this path with a given string. """
        with self.WriteContext(self.path, overwrite=overwrite) as write_path:
            return write_path.open_operation("w", lambda stream: stream.write(str(text)))

    def read(self, default=...):
        """ Read from this path to get a string.

            :rtype: str"""
        with self.ReadContext(self.path) as read_path:
            return read_path.open_operation("r", lambda stream: stream.read(), no_file_default=default)

    def append(self, text, newline=False):
        """ Append to this path with a given string.
            Optionally insert newline character before text. """
        if newline:
            text = f"\n{text}"

        with self.AppendContext(self.path) as append_path:
            return append_path.open_operation("a", lambda stream: stream.write(str(text)))

    def replace(self, d, regex=False):
        """ Replace matches in this path with a given dictionary. """
        with self.path.lock():
            text = self.path.text.read()
            for key, value in d.items():
                if regex:
                    text = re.sub(key, value, text)
                else:
                    text = text.replace(key, value)
            self.path.text.write(text, overwrite=True)











































import unittest
import multiprocessing as mp

from generalfile import *
from generalfile.test.setup_workdir import setup_workdir


def _thread_test(queue, i):
    queue.put(int(Path("test.txt").write(i, overwrite=True)))


class PathTest(unittest.TestCase):
    def setUp(self):
        """Set working dir and clear folder. Set path delimiter to '/' for testing."""
        setup_workdir()


class FileTest(PathTest):
    """ Skipped: open_folder, view, scrub"""
    def test_path(self):
        self.assertRaises(InvalidCharacterError, Path, "hello:there")
        self.assertRaises(InvalidCharacterError, Path, "hello<")
        self.assertRaises(InvalidCharacterError, Path, "hello>")
        self.assertRaises(InvalidCharacterError, Path, "hello.")

    def test_addPath(self):
        self.assertEqual(Path("foo/bar"), Path("foo") / "bar")
        self.assertEqual(Path("foo/bar"), Path("foo") / Path("bar"))
        self.assertEqual(Path("foo.txt/folder"), Path("foo.txt") / "folder")
        self.assertEqual(Path("folder/foo.txt"), Path("folder") / "foo.txt")

    def test_parts(self):
        path = Path("folder/folder2/test.txt")
        self.assertEqual(["folder", "folder2", "test.txt"], path.parts())

        self.assertEqual("foo", Path("foo/bar").parts()[0])
        self.assertEqual("bar", Path("foo/bar").parts()[1])

        self.assertEqual(not Path.verInfo.pathRootIsDelimiter, bool(Path().absolute().parts()[0]))

    def test_name(self):
        path = Path("folder/test.txt")
        self.assertEqual("test.txt", path.name())

        self.assertEqual("folder/foobar.txt", path.with_name("foobar.txt"))
        self.assertEqual("folder/hi", path.with_name("hi"))

    def test_stem(self):
        path = Path("folder/test.txt")
        self.assertEqual("test", path.stem())
        self.assertEqual("folder/foobar.txt", path.with_stem("foobar"))

        path = Path("folder/test.foo.txt.bar")
        self.assertEqual("test.foo.txt", path.stem())
        self.assertEqual("folder/foo.bar", path.with_stem("foo"))

    def test_true_stem(self):
        path = Path("folder/test.txt")
        self.assertEqual("test", path.true_stem())
        self.assertEqual("folder/foobar.txt", path.with_true_stem("foobar"))

        path = Path("folder/test.foo.txt.bar")
        self.assertEqual("test", path.true_stem())
        self.assertEqual("folder/yo.foo.txt.bar", path.with_true_stem("yo"))

    def test_suffixes(self):
        path = Path("folder/test.hello.txt")
        self.assertEqual([".hello", ".txt"], path.suffixes())
        self.assertEqual("folder/test.tsv", path.with_suffixes(".tsv"))

    def test_suffix(self):
        path = Path("folder/test")
        self.assertEqual("", path.suffix())

        path = Path("folder/test.txt")
        self.assertEqual(".txt", path.suffix())

        path = path.with_suffix("")
        self.assertEqual("folder/test", path)

        path = path.with_suffix(None)
        self.assertEqual("folder/test", path)

        path = path.with_suffix(".tsv")
        self.assertEqual("folder/test.tsv", path)

        path = path.with_suffix("")
        self.assertEqual("folder/test", path)

        path = path.with_suffix(".csv")
        self.assertEqual("folder/test.csv", path)

        path = path.with_suffix(".BACKUP", -2)
        self.assertEqual("folder/test.BACKUP.csv", path)

        path = path.with_suffix(".test", -2)
        self.assertEqual("folder/test.test.csv", path)

        path = path.with_suffix(None, 0)
        self.assertEqual("folder/test.csv", path)

        path = path.with_suffix(".foo", 2)
        self.assertEqual("folder/test.csv.foo", path)

        path = path.with_suffix(".bar", 3)
        self.assertEqual("folder/test.csv.foo.bar", path)

        path = path.with_suffix(".clamped", 5)
        self.assertEqual("folder/test.csv.foo.bar.clamped", path)

        path = path.with_suffix(".clamped", -10)
        self.assertEqual("folder/test.clamped.csv.foo.bar.clamped", path)

        path = path.with_suffix(None, 10)
        self.assertEqual("folder/test.clamped.csv.foo.bar", path)

        path = path.with_suffix(None, -10)
        self.assertEqual("folder/test.csv.foo.bar", path)

    def test_parent(self):
        path = Path("folder/foobar/test.txt")
        self.assertEqual(Path("folder/foobar"), path.get_parent())
        self.assertEqual(Path("folder/foobar"), path.get_parent(0))
        self.assertEqual(Path("folder"), path.get_parent(1, 1))
        self.assertEqual(Path(), path.get_parent(2, 2))
        self.assertEqual(None, path.get_parent(3, 3))
        self.assertEqual(None, path.get_parent(99, 99))
        self.assertEqual(None, path.get_parent(-99, -99))

        self.assertEqual([Path("folder/foobar"), Path("folder"), Path()], path.get_parents(depth=-1))
        new_path = Path("folder/foobar/test.txt")
        self.assertEqual([Path("folder/foobar"), Path("folder"), Path()], new_path.get_parents(depth=-1))

    def test_startswith(self):
        self.assertFalse(Path("file.txt").startswith("folder"))
        self.assertTrue(Path("file.txt").startswith("file"))
        self.assertFalse(Path("folder/file.txt").startswith("file.txt"))
        self.assertFalse(Path("folder/file.txt").absolute().startswith("folder"))

        self.assertTrue(Path("folder/file.txt").startswith("folder"))
        self.assertTrue(Path("file.txt").startswith("file.txt"))
        self.assertTrue(Path("file.SUFFIX.txt").startswith("file.SUFFIX.txt"))
        self.assertFalse(Path("filE.txt").startswith("file.txt"))

    def test_endswith(self):
        self.assertFalse(Path("file.txt").endswith("folder"))
        self.assertFalse(Path("file.txt").endswith("file"))
        self.assertFalse(Path("folder/file.txt").endswith("folder"))
        self.assertFalse(Path("folder/file.txt").absolute().endswith("file"))

        self.assertTrue(Path("folder/file.txt").endswith("file.txt"))
        self.assertTrue(Path("folder/file.txt").endswith("txt"))
        self.assertTrue(Path("file.txt").endswith("file.txt"))
        self.assertFalse(Path("filE.txt").endswith("file.txt"))

    def test_remove_start(self):
        self.assertEqual(Path(), Path("test.txt").remove_start("test.txt"))
        self.assertEqual(Path("folder/test.txt"), Path("folder/test.txt").remove_start("Folder"))
        self.assertEqual(Path("test.txt"), Path("folder/test.txt").remove_start("folder"))
        self.assertEqual(Path("folder/test.txt"), Path("folder/test.txt").remove_start("test"))

        if Path.verInfo.pathRootIsDelimiter:
            self.assertEqual(Path("test.txt"), Path("folder/test.txt").remove_start("folder"))


    def test_remove_end(self):
        self.assertEqual(Path(), Path("test.txt").remove_end("test.txt"))
        self.assertEqual(Path("test"), Path("test.txt").remove_end(".txt"), "test")
        self.assertEqual(Path("folder"), Path("folder/test.txt").remove_end("test.txt"))
        self.assertEqual(Path("folder/test.txt"), Path("folder/test.txt").remove_end("test"))

    def test_absolute(self):
        path = Path("test.txt")
        self.assertEqual(False, path.is_absolute())
        self.assertEqual(True, path.is_relative())

        path = path.absolute()
        self.assertEqual(True, path.is_absolute())
        self.assertEqual(False, path.is_relative())

        path = path.relative()
        self.assertEqual(False, path.is_absolute())
        self.assertEqual(True, path.is_relative())

        path = Path("folder/folder2/file.txt")
        self.assertEqual(Path("folder2/file.txt"), path.relative("folder"))
        self.assertEqual(path.relative("folder"), "folder2/file.txt")
        self.assertEqual(path.relative("folder/folder2"), "file.txt")

        self.assertEqual(path, path.relative("doesntexist"))

    def test_mirror_path(self):
        path = Path("foo")
        self.assertEqual(path.mirror_path().mirror_path(), path)
        self.assertEqual(True, path.mirror_path().is_absolute())

    def test_is_file_or_folder(self):
        Path("folder.txt/file.txt").write()
        self.assertEqual(True, Path("folder.txt").is_folder())
        self.assertEqual(False, Path("folder.txt").is_file())

        self.assertEqual(True, Path("folder.txt/file.txt").is_file())
        self.assertEqual(False, Path("folder.txt/file.txt").is_folder())

    def test_exists(self):
        path = Path("folder/test.txt")
        self.assertEqual(False, path.exists())
        self.assertEqual(False, Path("folder").exists())

        path.write()
        self.assertEqual(True, path.exists())
        self.assertEqual(True, Path("folder").exists())
        self.assertEqual(False, Path("folder/test").exists())

        Path("folder").delete()
        self.assertEqual(False, path.exists())
        self.assertEqual(False, Path("folder").exists())

    def test_working_dir(self):
        self.assertEqual(True, Path.get_working_dir().is_absolute())
        self.assertEqual(Path().absolute(), Path.get_working_dir())

        Path("folder").set_working_dir()
        self.assertEqual(True, Path.get_working_dir().endswith("folder"))
        self.assertEqual(Path().absolute(), Path.get_working_dir())

    def test_same_destination(self):
        path = Path("folder")
        self.assertEqual(True, path.same_destination(Path() / "folder"))
        self.assertEqual(True, path.same_destination(path.absolute()))
        self.assertEqual(True, path.same_destination("folder"))

    def test_write(self):
        self.assertEqual('"foobar"', Path("test.txt").write("foobar"))
        self.assertEqual("foobar", Path("test.txt").read())

        self.assertEqual('"foobar"', Path("test2").write("foobar"))
        self.assertEqual("foobar", Path("test2").read())

        self.assertEqual('"foobar"', Path("test2.doesntexist").write("foobar"))
        self.assertEqual("foobar", Path("test2.doesntexist").read())

        self.assertEqual('"foobar"', Path("folder/test.txt").write("foobar"))
        self.assertEqual("foobar", Path("folder/test.txt").read())

    def test_rename(self):
        Path("folder/test.txt").write()

        Path("folder/test.txt").rename(name="hello.txt")
        self.assertTrue(Path("folder/hello.txt").exists())
        self.assertFalse(Path("folder/test.txt").exists())

        Path("folder").rename(name="folder2")
        self.assertTrue(Path("folder2").exists())
        self.assertFalse(Path("folder").exists())

        Path("folder2/hello.txt").rename(name="foo.txt")
        self.assertTrue(Path("folder2/foo.txt").exists())

        Path("folder2/foo.txt").rename(name="foo.TEST.txt")
        self.assertTrue(Path("folder2/foo.TEST.txt").exists())

        Path("folder2/foo.TEST.txt").rename(name="foobar")
        self.assertTrue(Path("folder2/foobar").is_file())

        Path("folder2/foobar").rename(suffix=".test")
        self.assertTrue(Path("folder2/foobar.test").exists())

        self.assertEqual("folder2/hello.test", Path("folder2/foobar.test").rename(stem="hello"))
        self.assertTrue(Path("folder2/hello.test").exists())

    def test_copy(self):
        Path("folder/test.txt").write()
        Path("folder/test.txt").copy("foo.txt")
        self.assertEqual(True, Path("folder/foo.txt").exists())

        Path("folder").copy("new")
        self.assertEqual(True, Path("new/foo.txt").exists())

        Path("new/foo.txt").copy("new/bar.txt")
        self.assertEqual(True, Path("new/bar.txt").exists())

    def test_copy_to_folder(self):
        Path("folder/test.txt").write()
        Path("folder/test2.txt").write()

        Path("folder").copy_to_folder("folder2")
        self.assertEqual(True, Path("folder2/test.txt").exists())
        self.assertEqual(True, Path("folder2/test2.txt").exists())
        self.assertEqual(True, Path("folder/test2.txt").exists())

        Path("folder/test.txt").copy_to_folder("")
        self.assertEqual(True, Path("test.txt").exists())
        self.assertEqual(False, Path("test2.txt").exists())

        Path("folder").copy_to_folder(Path(), overwrite=True)
        self.assertEqual(True, Path("test2.txt").exists())

    def test_move(self):
        Path("folder/test.txt").write(5)
        Path("folder/test2.txt").write()

        Path("folder").move("folder2")
        self.assertEqual(False, Path("folder").exists())
        self.assertEqual(True, Path("folder2/test.txt").exists())
        self.assertEqual(True, Path("folder2/test2.txt").exists())

        Path("folder2/test.txt").move("")
        self.assertEqual(True, Path("test.txt").exists())
        self.assertEqual(False, Path("test2.txt").exists())
        self.assertEqual(False, Path("folder2/test.txt").exists())

        Path("folder/test.txt").write(2)
        with self.assertRaises(FileExistsError):
            Path("folder").move(Path())

        self.assertEqual(5, Path("test.txt").read())
        Path("folder").move(Path(), overwrite=True)
        self.assertEqual(2, Path("test.txt").read())

    def test_create_folder(self):
        path = Path("folder/folder2.txt")
        path.create_folder()

        self.assertEqual(True, path.is_folder())

    def test_trash_and_delete(self):
        for method in ("trash", "delete"):
            path = Path("file.txt")
            self.assertEqual(False, path.exists())
            self.assertEqual(False, getattr(path, method)())

            path.write()
            self.assertEqual(True, path.exists())
            self.assertEqual(True, getattr(path, method)())
            self.assertEqual(False, getattr(path, method)())

            path = Path("folder/file.txt")
            self.assertEqual(False, path.exists())
            self.assertEqual(False, getattr(path, method)())

            path.write()
            self.assertEqual(True, path.exists())
            self.assertEqual(True, getattr(path.get_parent(), method)())
            self.assertEqual(False, getattr(path.get_parent(), method)())
            self.assertEqual(False, Path("folder").exists())

    def test_trash_and_delete_folder_content(self):
        for method in ("trash_folder_content", "delete_folder_content"):
            setup_workdir()

            mainPath = Path("folder")
            path = mainPath / "file.txt"
            path2 = mainPath / "folder2/file2.txt"
            self.assertEqual(False, mainPath.exists())
            self.assertEqual(False, getattr(mainPath, method)())

            for targetPath in (mainPath, ):
                path.write()
                path2.write()
                self.assertEqual(True, getattr(targetPath, method)())
                self.assertEqual(False, getattr(targetPath, method)())
                self.assertEqual(True, mainPath.exists())
                self.assertEqual(False, path.exists())
                self.assertEqual(False, path2.exists())

    def test_get_paths(self):
        Path("test.txt").write()
        Path("folder/test2.txt").write()
        Path("folder/test3.txt").write()

        self.assertEqual(2, len(Path().get_children()))
        self.assertEqual(3, len(Path().get_children(include_self=True)))

        self.assertEqual(0, len(Path("test.txt").get_children()))
        self.assertEqual(1, len(Path("test.txt").get_children(include_self=True)))
        self.assertEqual(0, len(Path("test.txt").get_children()))

        self.assertEqual(4, len(Path().get_children(depth=3)))
        self.assertEqual(5, len(Path().get_children(depth=1, include_self=True)))
        self.assertEqual(5, len(Path().get_children(depth=-1, include_self=True)))
        self.assertEqual(3, len(Path().get_children(depth=0, include_self=True)))

        self.assertEqual(0, len(Path("folder/test2.txt").get_children(depth=-1)))

        self.assertEqual(["folder/test2.txt", "folder/test3.txt"], Path("folder").get_children())

    def test_time_created_and_modified(self):
        path = Path("test.txt")
        methods = (path.seconds_since_creation, path.seconds_since_modified)

        for method in methods:
            self.assertRaises(AttributeError, method)

        path.write()

        for method in methods:
            self.assertGreater(method(), 0)

        # Think you need to flush and stuff to make this work for windows atleast
        # self.assertEqual(methods[0](), methods[1]())
        # path.write("foobar", overwrite=True)
        # self.assertNotEqual(methods[0](), methods[1]())

    def test_getitem(self):
        self.assertEqual("f", Path("foobar")[0])
        self.assertEqual("fo", Path("foobar")[0:2])
        self.assertEqual("raboof", Path("foobar")[-1::-1])

    def test_iter(self):
        self.assertEqual(["f", "o", "o"], list(Path("foo")))
        self.assertIn("foo", Path("foobar"))

    def test_root(self):
        str_path = Path().absolute().get_parent(depth=-1, index=-1).path
        if Path.verInfo.pathRootIsDelimiter:
            self.assertEqual("/", str_path)
        else:
            self.assertTrue(len(str_path) == 3 and str_path[1] == ":" and str_path[2] == Path.path_delimiter)
        self.assertEqual(True, Path().absolute().get_parent(-1, -1).is_root())
        self.assertEqual(False, Path("foo").is_root())
        self.assertEqual(False, Path().absolute().is_root())

    def test_as_working_dir(self):
        working_dir = Path.get_working_dir()
        with Path("hello").as_working_dir():
            self.assertEqual(working_dir / "hello", Path.get_working_dir())

        self.assertEqual(working_dir, Path.get_working_dir())

    def test_match(self):
        self.assertEqual(True, Path("hello/there").match("The*"))
        self.assertEqual(True, Path("hello/there").match("*"))
        self.assertEqual(True, Path("hello/there").match("*h*"))
        self.assertEqual(True, Path(".git").match(".*"))
        self.assertEqual(True, Path(".git").match("."))
        self.assertEqual(True, Path("hello/there").match("The"))
        self.assertEqual(True, Path("hello/there").match("hello/there"))
        self.assertEqual(True, Path("foo/bar/hi").match("/bar/"))
        self.assertEqual(True, Path("foo/bar/hi").match("\\bar\\"))

        self.assertEqual(False, Path("hello/there").match("x"))
        self.assertEqual(False, Path("hello/there").match("*x*"))
        self.assertEqual(False, Path("hello/there").match("there/"))

    def test_encode(self):
        self.assertEqual("foo/bar", Path("foo\\bar").encode())
        self.assertEqual("foo/bar", Path("foo/bar").encode())
        self.assertEqual("foo%20bar", Path("foo bar").encode())
        self.assertEqual("foo/bar/hi%20there", Path("foo/bar\\hi there").encode())
        self.assertEqual("_hello/there_now.py", Path("_hello/there_now.py").encode())
        self.assertEqual("foo/_bar_now", Path("foo\\_bar_now").encode())

    def test_threads(self):
        threads = []
        queue = mp.Queue()
        count = 2
        for i in range(count):
            threads.append(mp.Process(target=_thread_test, args=(queue, i)))
        for thread in threads:
            thread.start()

        results = []
        for i in range(count):
            get = queue.get()
            self.assertNotIn(get, results)
            results.append(get)

        self.assertEqual(len(results), count)

    def test_CaseSensitivityError(self):
        Path("foo.txt").write("hi")
        # Path("foo.txt").get_parent()
        self.assertRaises(CaseSensitivityError, Path("Foo.txt").exists)

    def test_get_alternative_path(self):
        path = Path("foo/bar.txt")
        self.assertEqual(path, path.to_alternative().from_alternative())

        path = path.absolute()
        self.assertEqual(path, path.to_alternative().from_alternative())

    def test_get_cache_dir(self):
        self.assertEqual(True, Path.get_lock_dir().startswith(Path.get_cache_dir()))

    def test_lock(self):
        path = Path("foo.txt")
        with path.lock():
            self.assertEqual(True, path.get_lock_path().exists())

    def test_open_operation(self):
        path = Path("foo.txt")
        with path.lock():
            path.open_operation("w", lambda stream: stream.write("hi"))
            self.assertEqual("hi", path.open_operation("r", lambda stream: stream.read()))

    def test_size(self):
        path = Path("foo.txt")
        path.write("bar")
        self.assertEqual(True, path.size() > 1)

    def test_without_file(self):
        path = Path("foo/bar")
        self.assertEqual("foo/bar", path.without_file())
        path.write()
        self.assertEqual("foo", path.without_file())

    def test_get_differing_files(self):
        Path("one/bar").write("hey")
        Path("one/foo").write("hello")
        Path("two/foo").write("hi")

        for base, target in (("one", "two"), ("two", "one")):
            self.assertEqual({"bar"}, Path(base).get_differing_files(target, exist=True, content=False))
            self.assertEqual({"foo"}, Path(base).get_differing_files(target, exist=False, content=True))
            self.assertEqual({"foo", "bar"}, Path(base).get_differing_files(target, exist=True, content=True))
            self.assertEqual(set(), Path(base).get_differing_files(target, exist=False, content=False))

    def test_contains(self):
        path = Path("foo")
        path.text.write("hello there test")
        self.assertEqual(True, path.contains("there"))
        self.assertEqual(False, path.contains("hi"))

    def test_is_identical(self):
        Path("foo").write("hello")
        Path("bar").write("hello")
        self.assertEqual(True, Path("foo").is_identical("bar"))

        Path("bar").write("hi", overwrite=True)
        self.assertEqual(False, Path("foo").is_identical("bar"))

        Path("foo").write("hi\n", overwrite=True)
        self.assertEqual(False, Path("foo").is_identical("bar"))

    def test_empty(self):
        self.assertEqual(True, Path().empty())
        self.assertEqual(True, Path("new").empty())
        Path("new").create_folder()
        self.assertEqual(True, Path("new").empty())
        Path("new/file.txt").write("foo")
        self.assertEqual(False, Path("new").empty())
        self.assertEqual(False, Path("new/file.txt").empty())
        Path("new/file.txt").delete()
        self.assertEqual(True, Path("new").empty())

    def test_pack(self):
        Path("base/test.txt").write("hello")
        Path("base").pack("target")
        self.assertEqual(True, Path("target.zip").exists())
        Path("target").unpack("new")
        self.assertEqual("hello", Path("new/test.txt").read())

        Path("base").pack("target.tar.gz").unpack("tarnew")
        self.assertEqual("hello", Path("tarnew/test.txt").read())

        Path("base/folder/hi").write("there")
        Path("base").pack("packed/pack.zip").unpack("newbase")
        self.assertEqual("hello", Path("newbase/test.txt").read())
        self.assertEqual("there", Path("newbase/folder/hi").read())

    def test_recycle(self):
        self.assertIs(Path("hi/there"), Path("hi/there"))
        self.assertIs(Path("hi/there")._children, Path("hi/there")._children)
        self.assertIs(Path("hi/there"), Path("hi\\there"))
        self.assertIs(Path("hi/there"), Path("hi") / "there")
        self.assertIs(Path("hi\\there"), Path("hi") / "there")

        self.assertIsNot(Path("hithere"), Path("hi") / "there")
        self.assertIsNot(Path("hi"), Path("hi").absolute())

    def test_read_empty(self):
        with self.assertRaises(FileNotFoundError):
            Path("hey").read()
        self.assertEqual(None, Path("hi").read(default=None))








































def setup_workdir(use_working_dir=False):
    """ Class to set up working dir for tests, File extensions import this class.
        Set working dir and clear it after it's made sure it's correct path."""
    path = Path.get_working_dir() if use_working_dir else Path(__file__).get_parent()

    if path.endswith("test"):
        path /= "tests"
    else:
        path /= "test/tests"

    path.set_working_dir()

    if not Path.get_working_dir().endswith("test/tests"):
        raise EnvironmentError(f"Failed setting correct working dir, should be ..test/tests but it's {path}")
    path.delete_folder_content()


from generalfile.path import Path




from generallibrary.functions import wrapper_transfer


def deco_require_state(is_file=None, is_folder=None, exists=None, quick_exists=None):
    """ Decorator to easily configure and see which state to require. """
    def _decorator(function):
        def _wrapper(self, *args, **kwargs):
            """:param generalfile.Path self:"""
            if is_file is not None:
                if self.is_file() != is_file:
                    raise AttributeError(f"Path {self} is_file check didn't match ({is_file}).")
            elif is_folder is not None:
                if self.is_folder() != is_folder:
                    raise AttributeError(f"Path {self} is_folder check didn't match ({is_folder}).")
            elif exists is not None:
                if self.exists() != exists:
                    raise AttributeError(f"Path {self} exists check didn't match ({exists}).")
            elif quick_exists is not None:
                if self.exists() is not quick_exists:
                    raise AttributeError(f"Path {self} quick exists check didn't match ({quick_exists}).")
            return function(self, *args, **kwargs)
        return wrapper_transfer(function, _wrapper)
    return _decorator


def deco_preserve_working_dir(function):
    """ Decorator to preserve working dir if given function changes it somehow. """
    def _wrapper(*args, **kwargs):
        working_dir_path = args[0].Path.get_working_dir()
        result = function(*args, **kwargs)
        if working_dir_path != args[0].Path.get_working_dir():
            working_dir_path.set_working_dir()
        return result
    return wrapper_transfer(function, _wrapper)


def deco_return_if_removed(content):
    """ Decorator to return whether path exists to be removed or not. """
    def _decorator(function):
        def _wrapper(*args, **kwargs):
            path = args[0]
            if not path.exists():
                return False

            if content and path.empty():
                return False

            function(*args, **kwargs)
            return True
        return wrapper_transfer(function, _wrapper)
    return _decorator





from generalfile import Path
from generalfile.test.test_path import PathTest

import pandas as pd


class FileTest(PathTest):
    @classmethod
    def setUpClass(cls):
        """ . """
        import pandas
        cls.pd = pandas

    def test_tsvWriteAndRead(self):
        self._doTestsOnDataFrame({"a": {"color": "red", "value": 5}, "b": {"color": "blue", "value": 2}})
        self._doTestsOnDataFrame([{"color": "red", "value": 5}, {"color": "blue", "value": 2}])
        self._doTestsOnDataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
        self._doTestsOnDataFrame([[1, 2, 3], [4, 5, 6]])

    def test_tsvAppend(self):
        df = pd.DataFrame([[1, 2, 3], [4, 5, 6]])
        self._compareFrames(df, self._appendAndReadDF([[1, 2, 3], [4, 5, 6]]))
        self._compareFrames(df, self._appendAndReadDF([{"a": 1, "b": 2, "c": 3}, {"d": 4, "e": 5, "f": 6}]))
        self._compareFrames(df, self._appendAndReadDF({1: {"b": 2, "c": 3}, 4: {"e": 5, "f": 6}}))
        self._compareFrames(df, self._appendAndReadDF({1: [2, 3], 4: [5, 6]}))

    def _compareFrames(self, df1, df2):
        self.assertEqual(df1.shape, df2.shape)
        self.assertTrue(df1.columns.equals(df2.columns))
        self.assertTrue(df1.index.equals(df2.index))

    def _writeAndReadDF(self, df):
        header, column = Path("df.tsv").spreadsheet.write(df, overwrite=True)
        read = Path("df.tsv").spreadsheet.read(header=header, column=column)
        return read

    def _appendAndReadDF(self, obj):
        Path("df.tsv").spreadsheet.write(pd.DataFrame(), overwrite=True)
        Path("df.tsv").spreadsheet.append(obj)
        read = Path("df.tsv").spreadsheet.read(header=False, column=False)
        return read

    def _doTestsOnDataFrame(self, obj):
        df = pd.DataFrame(obj)
        self._compareFrames(df, self._writeAndReadDF(df))
        self._compareFrames(df.T, self._writeAndReadDF(df.T))

    def test_read_empty(self):
        with self.assertRaises(FileNotFoundError):
            Path("hey").spreadsheet.read()
        self.assertEqual(None, Path("hi").spreadsheet.read(default=None))





from setuptools import setup, find_namespace_packages
from pathlib import Path

try:
    long_description = (Path(__file__).parent / 'README.md').read_text(encoding='utf-8')
except FileNotFoundError:
    long_description = 'Readme missing'

setup(
    name="generalfile",
    author='Rickard "Mandera" Abraham',
    author_email="rickard.abraham@gmail.com",
    version="2.5.3",
    description="Easily manage files cross platform.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        'generallibrary',
        'send2trash',
        'appdirs',
        'pandas',
        'dill',
    ],
    url="https://github.com/ManderaGeneral/generalfile",
    license="mit",
    python_requires=">=3.8, <3.10",
    packages=find_namespace_packages(exclude=("build*", "dist*")),
    extras_require={},
    classifiers=[
        'Topic :: Desktop Environment :: File Managers',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
    ],
)




from generalfile import Path
from generalfile.test.test_path import PathTest


class FileTest(PathTest):
    def test_cfg(self):
        dict_ = {'test': {'foo': 'bar', 'number': 2, 'hi': ['a', 'b', 3]}}

        Path("foo").cfg.write(dict_)
        self.assertEqual(dict_, Path("foo").cfg.read())

        dict_["test"]["foo"] = "hi"
        Path("foo").cfg.write(dict_, overwrite=True)
        self.assertEqual(dict_, Path("foo").cfg.read())

    def test_cfg_append(self):
        dict_ = {'setup': {'hello': 'random'}}
        Path("foo").cfg.write(dict_)

        dict_ = {'setup': {'foo': 'random'}, "test": {5: "bar"}}
        Path("foo").cfg.append(dict_)

        self.assertEqual({'setup': {'foo': 'random'}, 'test': {'5': 'bar'}}, Path("foo").cfg.read())

    def test_read_empty(self):
        with self.assertRaises(FileNotFoundError):
            Path("hey").cfg.read()
        self.assertEqual(None, Path("hi").cfg.read(default=None))




from generallibrary import deco_cache
from generalfile.errors import CaseSensitivityError
from generalfile.decorators import deco_require_state, deco_preserve_working_dir, deco_return_if_removed

import pathlib
import appdirs
import os
import shutil
from send2trash import send2trash
import json
from distutils.dir_util import copy_tree
import time

from generalfile.optional_dependencies._extension import WriteContext, ReadContext


class Path_Operations:
    """ File operations methods for Path. """
    _suffixIO = {"plain_text": ("txt", "md", ""), "spreadsheet": ("tsv", "csv")}
    timeout_seconds = 5
    dead_lock_seconds = 3
    _working_dir = None

    def _removed_path(self):
        """ :param generalfile.Path self: """
        self.set_parent(None)

    def open_operation(self, mode, func, encoding=..., no_file_default=...):
        """ Handles all open() calls.

            :param generalfile.Path self:
            :param mode:
            :param func:
            :param encoding:
            :param no_file_default: """
        if encoding is Ellipsis:
            encoding = "utf-8"

        try:
            with open(self, mode, encoding=encoding) as stream:
                return func(stream)
        except FileNotFoundError as e:
            if no_file_default is ...:
                raise e
            else:
                return no_file_default

    def write(self, content=None, overwrite=False, indent=None):
        """ Write to this Path with JSON.

            :param generalfile.Path self:
            :param any content: Serializable by JSON
            :param overwrite: Whether to allow overwriting or not.
            :param indent: """
        content_json = "" if content is None else json.dumps(content, indent=indent)
        with WriteContext(self, overwrite=overwrite) as write_path:
            write_path.open_operation("w", lambda stream: stream.write(content_json))
        return content_json

    def read(self, default=...):
        """ Read this Path with JSON.

            :param generalfile.Path self:
            :param default: Optionally return a default value. """
        with ReadContext(self) as read_path:
            return read_path.open_operation("r", lambda stream: json.loads(stream.read()), no_file_default=default)

    @deco_require_state(exists=True)
    def rename(self, name=None, stem=None, suffix=None, overwrite=False):
        """ Rename this single file or folder to anything.

            :param generalfile.Path self:
            :param name:
            :param stem:
            :param suffix:
            :param overwrite: """
        new_path = self
        for key, value in {"stem": stem, "suffix": suffix, "name": name}.items():
            if value is not None:
                new_path = getattr(new_path, f"with_{key}")(value)
        if new_path == self:
            return

        with self.lock(new_path):
            if overwrite:
                self._path.replace(new_path)
            else:
                self._path.rename(new_path)
            self._removed_path()
        return new_path

    @deco_require_state(exists=True)
    def copy(self, new_path, overwrite=False):
        """ Copy a file or folder next to itself with a new name.
        If target exists then it is removed first, so it cannot add to existing folders, use `copy_to_folder` for that.

        :param generalfile.Path self:
        :param generalfile.Path or str new_path:
        :param overwrite:
        :return:
        """
        new_path = self.with_name(self.Path(new_path).name())

        with self.lock(new_path):
            if new_path.exists():
                if overwrite:
                    new_path.delete()
                else:
                    raise AttributeError(f"Target path '{new_path}' exists but overwrite is `False`.")
            self._copy_file_or_folder(new_path=new_path)

    def _copy_file_or_folder(self, new_path):
        """ :param generalfile.Path self: """
        if self.is_file():
            shutil.copy(self.path, new_path, follow_symlinks=False)  # Can clobber
        else:
            copy_tree(self.path, new_path.path)

    @deco_require_state(exists=True)
    def _copy_or_move(self, target_folder_path, overwrite, method):
        """ :param generalfile.Path self: """
        target_folder_path = self.Path(target_folder_path)
        if target_folder_path.is_file():
            raise NotADirectoryError("parent_path cannot be a file")

        self_parent_path = self.absolute().get_parent() if self.is_file() else self.absolute()
        if self_parent_path == target_folder_path:
            return

        if self.is_file():
            filepaths = [self]
        else:
            filepaths = self.get_children()

        target_filepaths = [target_folder_path / path.absolute().relative(self_parent_path) for path in filepaths]
        if not overwrite and any([target.exists() for target in target_filepaths]):
            raise FileExistsError("Atleast one target filepath exists, cannot copy")

        with self.lock(target_folder_path):
            target_folder_path.create_folder()
            for path, target in zip(filepaths, target_filepaths):

                if method == "copy":
                    self.__class__._copy_file_or_folder(path, target)  # Same as path._copy_file_or_folder(target)

                elif method == "move":
                    shutil.move(path, target)  # Can clobber if full target path is specified like we do

            if method == "move":
                self._removed_path()
                if self.is_folder():
                    self.delete()

    def copy_to_folder(self, target_folder_path, overwrite=False):
        """ Copy file or files inside given folder to anything except it's own parent, use `copy` for that.

            :param generalfile.Path self:
            :param target_folder_path:
            :param overwrite: """
        return self._copy_or_move(target_folder_path=target_folder_path, overwrite=overwrite, method="copy")

    def move(self, target_folder_path, overwrite=False):
        """ Move files inside given folder or file to anything except it's own parent.

            :param generalfile.Path self:
            :param target_folder_path:
            :param overwrite: """
        return self._copy_or_move(target_folder_path=target_folder_path, overwrite=overwrite, method="move")

    def is_file(self):
        """ Get whether this Path is a file.

            :param generalfile.Path self: """
        return self._path.is_file()

    def is_folder(self):
        """ Get whether this Path is a folder.

            :param generalfile.Path self: """
        return self._path.is_dir()

    def is_root(self):
        """ Get whether this Path is a root.

            :param generalfile.Path self: """
        if self.verInfo.pathRootIsDelimiter:
            return self.path == self.path_delimiter
        else:
            return len(self.path) == 3 and self.path[1] == ":" and self.path[2] == self.path_delimiter

    def _case_sens_test(self, path):
        """ :param generalfile.Path self: """
        return self != path and self.path.lower() == str(path).lower()

    def exists(self):
        """ Get whether this Path exists.

            :param generalfile.Path self: """
        if self.get_sibling(spawn=False, filt=self._case_sens_test, traverse_excluded=True):
            raise CaseSensitivityError(f"Same path with differing case not allowed: '{self}'")
        return self._path.exists()

    def empty(self):
        """ Get whether path is an empty folder or not.

            :param generalfile.Path self: """
        if not self.exists():
            return True
        elif self.is_file():
            return False
        if self.get_child():
            return False
        return True

    def without_file(self):
        """ Get this path without it's name if it's a file, otherwise it returns itself.

            :param generalfile.Path self: """
        if self.is_file():
            return self.get_parent()
        else:
            return self

    def create_folder(self):
        """ Create folder with this Path unless it exists. 

            :param generalfile.Path self: """
        if self.exists():
            return False
        else:
            self._path.mkdir(parents=True, exist_ok=True)
            return True

    def open_folder(self):
        """ Open folder to view it manually.

            :param generalfile.Path self: """
        self.create_folder()
        os.startfile(self.without_file())

    @classmethod
    def get_working_dir(cls):
        """ Get current working folder as a new Path.
            Falls back to last seen working_dir if it doesn't exist. (Only seems to raise Error on posix)

            :param generalfile.Path cls:
            :rtype: generalfile.Path """
        # return cls.Path(pathlib.Path.cwd())

        try:
            working_dir = cls.Path(pathlib.Path.cwd())
        except FileNotFoundError as e:
            if cls._working_dir is None:
                raise e
            else:
                return cls._working_dir
        else:
            cls._working_dir = working_dir
            return working_dir

    def set_working_dir(self):
        """ Set current working folder.

            :param generalfile.Path self: """
        self.create_folder()
        self._working_dir = self.absolute()
        os.chdir(self._working_dir)

    @classmethod
    @deco_cache()
    def get_cache_dir(cls):
        """ Get cache folder.

            :param generalfile.Path or Any cls:
            :rtype: generalfile.Path """
        path = cls(appdirs.user_cache_dir())
        return path

    @classmethod
    @deco_cache()
    def get_lock_dir(cls):
        """ Get lock folder inside cache folder.

            :param generalfile.Path cls:
            :rtype: generalfile.Path """
        return cls.get_cache_dir() / "generalfile" / "locks"

    @deco_cache()
    def get_lock_path(self):
        """ Get absolute lock path pointing to actual lock.

            :param generalfile.Path self:
            :rtype: generalfile.Path """
        return self.get_lock_dir() / self.absolute().to_alternative()

    @deco_preserve_working_dir
    @deco_return_if_removed(content=False)
    def delete(self, error=True):
        """ Delete a file or folder.

            :param error:
            :param generalfile.Path self: """
        with self.lock():
            try:
                if self.is_file():
                    os.remove(self)
                elif self.is_folder():
                    shutil.rmtree(self)
            except PermissionError:
                self.trash()  # Sometimes failing to remove a git file on windows, send2trash still worked
            except Exception as e:
                if error:
                    raise e

            self._removed_path()

    @deco_preserve_working_dir
    @deco_return_if_removed(content=False)
    def trash(self):
        """ Trash a file or folder.

            :param generalfile.Path self: """
        with self.lock():
            send2trash(self.path)

            self._removed_path()

    @deco_preserve_working_dir
    @deco_return_if_removed(content=True)
    def delete_folder_content(self):
        """ Delete every path in a folder.

            :param generalfile.Path self: """
        for path in self.get_children(gen=True):
            path.delete()

    @deco_preserve_working_dir
    @deco_return_if_removed(content=True)
    def trash_folder_content(self):
        """ Trash a file or folder and then create an empty folder in it's place.

            :param generalfile.Path self: """
        for path in self.get_children(gen=True):
            path.trash()

    @deco_require_state(is_file=True)
    def seconds_since_creation(self):
        """ Get time in seconds since file was created.
            NOTE: Doesn't seem to update very quickly for windows (7).

            :param generalfile.Path self: """
        return time.time() - os.path.getctime(self)

    @deco_require_state(is_file=True)
    def seconds_since_modified(self):
        """ Get time in seconds since file was modified.

            :param generalfile.Path self: """
        return time.time() - os.path.getmtime(self)

    @deco_require_state(is_file=True)
    def size(self):
        """ Get size in bytes of file.

            :param generalfile.Path self: """
        return self._path.stat().st_size

    def is_identical(self, path):
        """ Get whether this file's content is identical to another.

            :param generalfile.Path self:
            :param path: """
        path = self.Path(path)
        self_exists = self.exists()
        path_exists = path.exists()

        if not self_exists or not path_exists:
            return self_exists == path_exists

        with self.lock(path):
            with open(self, "r") as file1:
                with open(path, "r") as file2:
                    return file1.read() == file2.read()

    @deco_require_state(is_folder=True)
    def get_differing_files(self, target, exist=True, content=True, filt=None):
        """ Get a set of changed files by comparing two folders.

            :param generalfile.Path self:
            :param target:
            :param exist: Whether to compare files' existence. Ignores content.
            :param content: Whether to compare files' content if both files exist.
            :param filt: Optional filter, takes one Path as arg. """
        target = self.Path(target)
        assert target.is_folder()

        if filt is None:
            new_filt = lambda path: path.is_file()
        else:
            new_filt = lambda path: path.is_file() and filt(path)

        self_paths = {child.relative(self) for child in self.get_children(depth=-1, filt=new_filt, traverse_excluded=True)}
        target_paths = {child.relative(target) for child in target.get_children(depth=-1, filt=new_filt, traverse_excluded=True)}

        diff = set()
        if exist:
            diff.update(self_paths.symmetric_difference(target_paths))
        if content:
            diff.update({path for path in self_paths.intersection(target_paths) if path not in diff and not (self / path).is_identical(path=target / path)})
        return diff

    @deco_require_state(is_file=True)
    def contains(self, text):
        """ Return whether text string exists in one of the lines in this file.

            :param generalfile.Path self:
            :param text: """
        with self.lock():
            with open(self, "r") as stream:
                for line in stream:
                    if text in line:
                        return True
        return False

    def _pack_default_suffix(self):
        """ :param generalfile.Path self: """
        if not self.suffix():
            return self.with_suffix(".zip")
        else:
            return self

    @deco_require_state(is_folder=True)
    def pack(self, target, overwrite=False):
        """ Pack self which is folder to a new target archive.

            :param generalfile.Path self: Base folder to be packed.
            :param target: Full path of new archive. Optional suffix, defaults to zip if missing.'
            :param overwrite: """
        target = self.Path(target)._pack_default_suffix().absolute()
        if not overwrite:
            assert not target.exists()

        root_dir = self.absolute()
        target_stem = str(target.with_suffixes())
        target_suffix = "".join(target.suffixes())[1:]
        target_suffix = {"tar.gz": "gztar"}.get(target_suffix, target_suffix)

        shutil.make_archive(root_dir=root_dir, base_name=target_stem, format=target_suffix)

        return target

    def unpack(self, base, overwrite=False):
        """ Unpack self which is archive to target folder (Must be empty if overwrite is False).

            :param generalfile.Path self:
            :param base:
            :param overwrite: """
        base = self.Path(base)
        if not overwrite:
            assert not base.exists() or base.empty()

        shutil.unpack_archive(filename=self._pack_default_suffix(), extract_dir=base)
        return base



















